% IMPLEMENTACAO DA INTERFACE COM O USUARIO %

functor

import
   Application
   System
   Open
   Resolvedor

define

   % Procedimentos implementados:
   % LeLinha, LeEntrada, GravaSolucao, ResolveGravando, VerificaInterrupcao.

   % Le uma linha do arquivo dado e a devolve, desprezando um eventual
   % caractere final de quebra de linha.
   fun {LeLinha Arquivo}
      {String.token {Arquivo getS($)} &\r $ _}
   end

   % Le, a partir do arquivo ArqEntrada, todos os dados de entrada do programa.
   % Os dados sao guardados num registro, associado a variavel Entrada.
   proc {LeEntrada ArqEntrada Entrada}
      % Este procedimento implementa os seguintes procedimentos auxiliares:
      % LeTurma, LeRestrProf, LeGrupo, LePreferencia.

      % Despreza a linha inicial
      {LeLinha ArqEntrada _}
      % Obtem o numero de dias
      NumDias = {String.toInt {LeLinha ArqEntrada}}
      % Despreza uma linha
      {LeLinha ArqEntrada _}
      % Obtem o numero de turmas
      NumTurmas = {String.toInt {LeLinha ArqEntrada}}
      % Cria o dicionario de turmas
      Turmas = {Dictionary.new}
      % Numero de professores que tem restricoes de horarios
      NumRestrProf
      % Lista de restricoes de horarios dos professores
      RestrProfessores
      % Numero de grupos de exclusao mutua
      NumGrupos
      % Lista de grupos de exclusao mutua
      Grupos
      % Numero de preferencias
      NumPreferencias
      % Lista de preferencias
      Preferencias

      % Le uma turma do arquivo de entrada e a coloca no dicionario de turmas
      proc {LeTurma Turmas}
         % Este procedimento implementa o seguinte procedimento auxiliar:
         % LeAula.

         % Le a identificacao da turma
         IdTurma = {String.toAtom {LeLinha ArqEntrada}}
         % Le o nome do professor da turma
         Prof = {String.toAtom {LeLinha ArqEntrada}}
         % Le o "flag" (0 ou 1) que indica se esta turma tem restricoes de
         % exclusao mutua com outra(s).
         ExisteExclMut = {String.toInt {LeLinha ArqEntrada}}
         % A lista de exclusao mutua desta turma (nil, se nao houver
         % exclusao mutua).
         ListaExclMut
         % O numero de aulas desta turma
         NumAulas
         % A lista de aulas desta turma
         ListaAulas
         % A distancia minima entre as aulas desta turma
         DistMin
         % A distancia maxima entre as aulas desta turma
         DistMax
         % true, se os horarios das aulas desta turma devem ser diferentes.
         % Caso contrario, false.
         HorVaria

         % Le uma aula do arquivo de entrada e devolve o registro equivalente
         % a aula lida
         fun {LeAula}
            % Le a duracao da aula
            Duracao = {String.toInt {LeLinha ArqEntrada}}
            % Leitura dos horarios %
            ListaHorarios =
            {List.map {String.tokens {LeLinha ArqEntrada} &  $}
             fun {$ HorString}
                [DS HS MS] = {String.tokens HorString &: $}
                D = {String.toInt DS}
                H = {String.toInt HS}
                M = {String.toInt MS}
             in
                D#H#M
             end
            }
         in
            aula(dur:Duracao hor:ListaHorarios)
         end
         % FIM DO PROCEDIMENTO LeAula %%

      in
         % Leitura da lista de exclusao mutua %
         if ExisteExclMut == 1 then
            ListaExclMut =
            {List.map { String.tokens {LeLinha ArqEntrada} &  $}
             fun {$ TurmaString}
                {String.toAtom TurmaString}
             end
            }
         else
            ListaExclMut = nil
         end
         % Leitura das aulas %
         NumAulas = {String.toInt {LeLinha ArqEntrada}}
         {List.foldL {List.number 1 NumAulas 1}
          fun {$ ListaAtual _}
             { List.append ListaAtual [{LeAula}] $}
          end
          nil ListaAulas
         }
         % Leitura das distancias %
         [DistMin DistMax] =
         {List.map {String.tokens {LeLinha ArqEntrada} &  $}
          fun {$ DistString}
             {String.toInt DistString}
          end
         }
         % Leitura do parametro que indica se os horarios das aulas devem
         % variar %
         if {String.toInt {LeLinha ArqEntrada}} == 1 then
            HorVaria = true
         else
            HorVaria = false
         end
         % Coloca a turma lida no dicionario
         {Dictionary.put Turmas IdTurma
          turma(prof:Prof exclMut:ListaExclMut aulas:ListaAulas
                dist:dist(min:DistMin max:DistMax) horVaria:HorVaria)}
      end
      % FIM DO PROCEDIMENTO LeTurma %%

      % Le as restricoes de horarios de um professor e devolve o registro
      % correspondente.
      fun {LeRestrProf}
         % Le o nome do professor
         Prof = {String.toAtom {LeLinha ArqEntrada}}
         % Leitura dos horarios impossiveis %
         HorariosImpossiveis =
         {List.map {String.tokens {LeLinha ArqEntrada} &  $}
          fun {$ HorString}
             [DS HIS HFS] = {String.tokens HorString &: $}
             D = {String.toInt DS}
             HI = {String.toInt HIS}
             HF = {String.toInt HFS}
          in
             D#HI#HF
          end
         }
      in
         restrProf(prof:Prof impossiveis:HorariosImpossiveis)
      end
      %% FIM DO PROCEDIMENTO LeRestrProf %%

      % Le um grupo de exclusao mutua e devolve a lista correspondente.
      fun {LeGrupo}
         % Le a lista de turmas
         {List.map {String.tokens {LeLinha ArqEntrada} &  $}
          fun {$ TurmaString}
             {String.toAtom TurmaString $}
          end
          $
         }
      end
      %% FIM DO PROCEDIMENTO LeGrupo %%
      
      % Le uma preferencia e devolve o registro correspondente.
      fun {LePreferencia}
         % Le o nome do professor
         Prof = {String.toAtom {LeLinha ArqEntrada}}
         % Leitura dos horarios indesejados %
         Indesejados =
         {List.map {String.tokens {LeLinha ArqEntrada} &  $}
          fun {$ HorString}
             [DS HIS HFS] = {String.tokens HorString &: $}
             D = {String.toInt DS}
             HI = {String.toInt HIS}
             HF = {String.toInt HFS}
          in
             D#HI#HF
          end
         }
         % Leitura do peso da preferencia
         Peso = {String.toInt {LeLinha ArqEntrada}}
      in
         preferencia(prof:Prof indesejados:Indesejados peso:Peso)
      end
      %% FIM DO PROCEDIMENTO LePreferencia %%
      
   in
      % OBTENCAO DAS TURMAS %
      {System.showInfo "Carregando as turmas..."}
      {For 1 NumTurmas 1
       proc {$ _}
          {LeTurma Turmas}
       end
      }
      % OBTENCAO DAS RESTRICOES DE HORARIOS DOS PROFESSORES %
      {LeLinha ArqEntrada _}
      {System.showInfo "Carregando as restricoes dos professores..."}
      NumRestrProf = {String.toInt {LeLinha ArqEntrada}}
      if NumRestrProf > 0 then
         {List.foldL {List.number 1 NumRestrProf 1}
          fun {$ ListaAtual _}
             {List.append ListaAtual [{LeRestrProf}] $}
          end
          nil RestrProfessores
         }
      else RestrProfessores = nil end
      % OBTENCAO DOS GRUPOS DE EXCLUSAO MUTUA %
      {LeLinha ArqEntrada _}
      {System.showInfo "Carregando os grupos de exclusao mutua..."}
      NumGrupos = {String.toInt {LeLinha ArqEntrada}}      
      if NumGrupos > 0 then
         {List.foldL {List.number 1 NumGrupos 1}
          fun {$ ListaAtual _}
             {List.append ListaAtual [{LeGrupo}] $}
          end
          nil Grupos
         }
      else Grupos = nil end
      % OBTENCAO DAS PREFERENCIAS %
      {LeLinha ArqEntrada _}
      {System.showInfo "Carregando as preferencias para otimizacao..."}
      NumPreferencias = {String.toInt {LeLinha ArqEntrada}}
      if NumPreferencias > 0 then
         {List.foldL {List.number 1 NumPreferencias 1}
          fun {$ ListaAtual _}
             {List.append ListaAtual [{LePreferencia}] $}
          end
          nil Preferencias
         }
      else Preferencias = nil end

      % Cria um registro com os dados de entrada do problema
      Entrada = entrada(numDias:NumDias turmas:Turmas
                        restrProf:RestrProfessores grupos:Grupos
                        preferencias:Preferencias)

   end
   %% FIM DO PROCEDIMENTO LeEntrada %%

   % Recebe:
   % - Uma solucao para o problema.
   % - Um arquivo
   % Grava a solucao dada no arquivo dado
   proc {GravaSolucao Solucao Arquivo}
      % Implementa os seguintes procedimentos auxiliares:
      % Soma, FormataHorario.

      % Recebe:
      % - Um horario, na forma Hora#Minuto.
      % - Uma quantidade em minutos.
      % Devolve: a soma das duas entradas, na forma Hora#Minuto.
      fun {Soma H#M Min}
         SomaEmMinutos = H*60 + M + Min
      in
         (SomaEmMinutos div 60)#(SomaEmMinutos mod 60)
      end

      % Recebe um horario, na forma Hora#Minuto, e devolve devolve uma String,
      % na forma "XhYmin".
      fun {FormataHorario H#M}
         if M < 10 then
            ""#H#"h0"#M#"min"
         else
            ""#H#"h"#M#"min"
         end
      end

      % Lista de aulas, estando cada uma na forma:
      % Turma#Dia#H#M#Duracao
      AulasParaImpressao
      % Uma tupla (para todo dia I, todas as aulas do dia I estarao na
      % posicao I da tupla).
      AulasPorDia
   in
      {Arquivo putS("*** DETALHES DA SOLUCAO ***\n")}
      % Imprime o numero de falhas da solucao %
      {Arquivo putS("Custo (soma das preferencias nao satisfeitas): "#Solucao.falhas#"\n")}
      % Imprime as preferencias violadas %
      {Arquivo putS("* Lista de preferencias violadas *")}
      {List.forAll Solucao.violadas
       proc {$ Violada}
          {Arquivo putS("Professor: "#Violada.prof)}
          {Arquivo putS("Peso: "#Violada.peso)}
          {Arquivo putS("Aulas que violam a preferencia:")}
          {List.forAll Violada.aulas
           proc {$ Aula}
              HF#MF = {Soma Aula.hor.1#Aula.hor.2 Aula.dur}
           in
              {Arquivo putS("Turma "#Aula.turma#", no dia "#Aula.dia#", de "#
                            {FormataHorario Aula.hor}#" a "#
                            {FormataHorario HF#MF})}
           end
          }
          {Arquivo putS("----------------------------")}
       end
      }
      % Cria uma lista de aulas, estando cada aula na forma:
      % Turma#Dia#H#M#Duracao
      AulasParaImpressao =
      {List.foldL {Dictionary.entries Solucao.turmas}
       fun {$ ListaAtual IdTurma#ListaDeAulas}
          {List.foldL ListaDeAulas
           fun {$ Atual Aula}
              {List.append Atual
               [IdTurma#Aula.dia#Aula.hor.1#Aula.hor.2#Aula.dur] $}
           end
           ListaAtual $
          }
       end
       nil $
      }
      % Separa as aulas por dia %
      AulasPorDia = {Tuple.make aulas DadosEntrada.numDias}
      {List.foldL {List.number 1 DadosEntrada.numDias 1}
       proc {$ Restante Dia Prox}
          AulasDesteDia =
          {List.partition Restante
           fun {$ _#D#_#_#_}
              D == Dia
           end
           $ Prox
          }
       in
          % Ordena as aulas por hora e minuto %
          AulasPorDia.Dia =
          {List.sort AulasDesteDia
           fun {$ _#_#H1#M1#_ _#_#H2#M2#_}
              H1 < H2 orelse (H1 == H2 andthen M1 =< M2)
           end
          }
       end
       AulasParaImpressao _
      }
      % Impressao das aulas %
      {For 1 DadosEntrada.numDias 1
       proc {$ Dia}
          {Arquivo putS("** Aulas do dia "#Dia#" **")}
          {List.forAll AulasPorDia.Dia
           proc {$ T#_#H#M#Dur}
              HF#MF = {Soma H#M Dur}
           in
              {Arquivo putS(T#": de "#{FormataHorario H#M}#" a "#
                            {FormataHorario HF#MF}#".")}
           end
          }
          {Arquivo putS("")}
       end
      }
      {Arquivo putS("----------------------------")}
      {Arquivo putS("----------------------------")}
      {Arquivo putS("")}
   end
   %% FIM DO PROCEDIMENTO GravaSolucao %%

   % Recebe:
   % - O resolvedor para o problema.
   % - Um arquivo.
   % Devolve: uma solucao otima para o problema (numa lista de um elemento).
   % Grava as solucoes encontradas (inclusive a otima) no arquivo Arquivo.
   fun {ResolveGravando RegistroResolvedor Arquivo}
      % Implementa o seguinte procedimento auxiliar:
      % ResolveComSolucao.

      % Recebe:
      % - Uma solucao (nil, se ainda nao ha solucoes), numa lista de um
      % elemento.
      % - A duracao acumulada da busca.
      % Devolve uma solucao otima (numa lista de um elemento), melhor que a
      % solucao dada (caso nao exista, devolve nil).
      fun {ResolveComSolucao SolucaoAtual DuracaoAtual}
         TI = {Time.time}
         Solucao = {RegistroResolvedor.proximaSolucao}
         % Tempo (em segundos) gasto na busca da proxima solucao.
         T = {Time.time} - TI
         % Nova duracao acumulada.
         Duracao = T + DuracaoAtual
      in
         % Verifica se a busca terminou.
         if Solucao == nil then
            DurMin = Duracao div 60
            DurSeg = Duracao mod 60
            % Numero de segundos (com um zero a esquerda, se menor que 10).
            SegString =
            if DurSeg < 10 then
               "0"#DurSeg
            else
               DurSeg
            end
         in
            {System.showInfo "Fim da busca, em "#T#" segundo(s)."}
            {System.showInfo "Duracao total: "#DurMin#"min"#SegString#"s."}
            SolucaoAtual
         else  % Foi encontrada outra solucao
            SolucaoEmSi = Solucao.1
         in
            {System.printInfo
             "Foi encontrada uma solucao de custo "#SolucaoEmSi.falhas#
             ", em "#T#" segundo(s). "}
            % Grava a nova solucao encontrada.
            {GravaSolucao SolucaoEmSi Arquivo}
            % Procura solucao melhor, se isso fizer sentido.
            if SolucaoEmSi.falhas > 0 then
               {System.showInfo "Buscando uma solucao melhor..."}
               {ResolveComSolucao Solucao Duracao}
            else
               DurMin = Duracao div 60
               DurSeg = Duracao mod 60
               % Numero de segundos (com um zero a esquerda, se menor que 10).
               SegString =
               if DurSeg < 10 then
                  "0"#DurSeg
               else
                  DurSeg
               end
            in
               {System.showInfo "Fim da busca."}
               {System.showInfo "Duracao total: "#DurMin#"min"#SegString#"s."}
               Solucao
            end
         end
      end
   in
      {ResolveComSolucao nil 0}
   end
   %% FIM DO PROCEDIMENTO ResolveGravando %%


   %%% EXECUCAO DA APLICACAO %%%


   % OBTEM OS ARGUMENTOS DE LINHA DE COMANDO %

   % Opcoes devem ser: --ae [arquivo de entrada] --as [arquivo de saida]
   Argumentos = {Application.getArgs record('ae'(single type:string)
                                            'as'(single type:string))}

   % CLASSE PARA LIDAR COM ARQUIVOS TEXTO
   class ArquivoTexto from Open.file Open.text end

   % CRIACAO DE UM ARQUIVO PARA A ENTRADA PADRAO %

   EntradaPadrao = { New ArquivoTexto
                     init(name:stdin flags:[read text]) }

   % OBTENCAO DOS DADOS DE ENTRADA DO PROBLEMA %

   % Abertura do arquivo de entrada
   ArqEntrada = { New ArquivoTexto
                  init(name:Argumentos.'ae' flags:[read text]) }

   % Leitura dos dados de entrada
   DadosEntrada
   {LeEntrada ArqEntrada DadosEntrada}
   {ArqEntrada close}
   {System.showInfo "Dados de entrada carregados com sucesso."}

   % RESOLUCAO DO PROBLEMA %

   % Exibicao das opcoes ao usuario
   {System.showInfo "\nEscolha uma das seguintes opcoes:"}
   {System.showInfo
    "1- Tentar antecipar o aparecimento da solucao otima (esta opcao pode diminuir muito a quantidade de solucoes encontradas antes da solucao otima)."}
   {System.showInfo
    "2- O seu interesse esta em qualquer solucao (esta opcao pode retardar o aparecimento da solucao otima)."}
   % Obtencao da opcao do usuario
   Opcao = {LeLinha EntradaPadrao}

   % Criacao de um resolvedor para o problema
   RegResolvedor = {Resolvedor.criaResolvedor DadosEntrada.numDias
                    DadosEntrada.turmas DadosEntrada.grupos
                    DadosEntrada.preferencias DadosEntrada.restrProf
                    if Opcao == "1" then otima else todas end
                   }

   % Abertura do arquivo de saida
   ArqSaida = { New ArquivoTexto
                init(name:Argumentos.'as' flags:[create text write truncate]) }

   % Procedimento auxiliar:
   % Le a entrada padrao, ate que o usuario digite "sair<Enter>".
   % Se isso acontecer, a aplicacao e encerrada (de forma adequada).
   proc {VerificaInterrupcao}
      {Delay 1000}
      Linha = {LeLinha EntradaPadrao}
   in
      if Linha == "sair" orelse Linha == "SAIR" then
         % Para a busca, se ela estiver em curso
         {RegResolvedor.pausa}
         % Fecha a entrada padrao, assim como os arquivos de entrada e de saida
         {EntradaPadrao close}
         {ArqEntrada close}
         {ArqSaida close}
         % Encerra a aplicacao
         {Application.exit 0}
      else
         {Delay 1000}
         {VerificaInterrupcao}
      end
   end

   % Inicia uma thread para verificar a entrada padrao (e encerrar a aplicacao)
   % se o usuario digitar sair<Enter>.
   thread {VerificaInterrupcao} end

   % Chama o coletor de lixo, antes de comecar a resolver o problema
   {System.gcDo}

   % A solucao do problema, numa lista de um elemento (ou nil, se nao existir
   % solucao).
   ListaComSolucao = {ResolveGravando RegResolvedor ArqSaida}

   if ListaComSolucao == nil then
      {System.showInfo "ATENCAO: o problema especificado nao tem solucao."}
   end

   % Fecha a entrada padrao
   {EntradaPadrao close}
   % Fecha o arquivo de saida
   {ArqSaida close}
   % Encerra a aplicacao
   {Application.exit 0}

end