% Este arquivo contem as funcoes que de fato resolvem o problema
functor

import
   FD
   Schedule
   Space
   Adaptador

export
   % Funcao que devolve um script para resolver a instancia dada do problema
   ScriptHorarios
   % Procedimento para ordenar solucoes, permitindo a busca da solucao otima
   Ordem

define
   % Sao implementados os seguintes procedimentos:
   % ScriptHorarios, Ordem.

   % Peso maximo que uma preferencia pode ter
   PesoMaximo = 5

   % Recebe:
   % Os dados de entrada para uma instancia de problema.
   %
   % Devolve um procedimento capaz de resolver o problema especificado pelos
   % dados de entrada.
   fun {ScriptHorarios NumDias TurmasEntrada GruposTurmas PreferenciasEntrada
        RestrProfEntrada Foco}
      % E implementado os seguinte procedimento auxiliar:
      % Falhas.

      % Tamanho (em minutos) da unidade de tempo utilizada pelo script
      TamanhoUnidade = 5
      % O numero de unidades (de TamanhoUnidade minutos) em um dia
      UnidadesPorDia = (24*60) div TamanhoUnidade
      % DADOS DE ENTRADA NUMA FORMA MAIS ADEQUADA PARA O SCRIPT
      % Tupla de aulas
      Aulas
      % Numero total de aulas
      NumAulas
      % Dicionario de turmas
      Turmas
      % Grupos de aulas para exclusao mutua
      Grupos
      % Dicionario de professores, contendo aulas
      Professores
      % Lista de preferencias
      Preferencias
      % Numero de preferencias
      NumPreferencias
      % Lista de restricoes de horarios dos professores
      RestrProfessores
      % Matriz de exclusao mutua entre as aulas
      MatrizExclMut

      % Recebe:
      % - Uma tupla Inicios (horarios de inicio das aulas (em unidades desde o
      % inicio do dia 1).
      % - Uma tupla de Duracoes, contendo as duracoes das aulas em unidades.
      %
      % Cria propagadores para a restricao de que cada elemento da lista
      % NaoSatisfeitas corresponde a uma preferencia (na mesma ordem da lista
      % Preferencias), da seguinte forma:
      % Se a preferencia for satisfeita, a posicao correspondente na lista
      % NaoSatisfeitas recebera 0. Caso contrario, a posicao recebera o peso
      % da preferencia.
      proc {Falhas Inicios Duracoes NaoSatisfeitas}
         NaoSatisfeitas = {FD.list {List.length Preferencias} 0#PesoMaximo}
         % Percorre as preferencias...
         {List.forAllInd Preferencias
          proc {$ IndPref Preferencia}
             % Aulas do professor que tem esta preferencia
             AulasDoProf = {Dictionary.get Professores Preferencia.prof}
             Indesejados = Preferencia.indesejados
             Peso = Preferencia.peso
             % Indica se alguma das aulas deste professor tem interseccao nao
             % nula com algum dos intervalos indesejados
             AlgumaNaoSatisfaz
          in
             % Cria os propagadores adequados para o valor de AlgumaNaoSatisfaz
             {List.foldL AulasDoProf
              fun {$ Atual NumAula}
                 {List.foldL Indesejados
                  fun {$ ResultAtual I#F}
                     {FD.disj ResultAtual
                      {FD.conj Inicios.NumAula <: F
                       Inicios.NumAula+Duracoes.NumAula >: I} $}
                  end
                  Atual $
                 }
              end
              0 AlgumaNaoSatisfaz
             }
             % Associa a posicao correspondente da lista NaoSatisfeitas
             % ao valor de AlgumaNaoSatisfaz para a preferencia corrente.
             {List.nth NaoSatisfeitas IndPref} =: AlgumaNaoSatisfaz*Peso
          end
         }
      end
      %% FIM DO PROCEDIMENTO Falhas %%

      % Adapta a entrada para um formato mais adequado ao script
      [Aulas Turmas Grupos Professores Preferencias RestrProfessores
       MatrizExclMut] =
      {Adaptador.adaptaAoScript NumDias TurmasEntrada GruposTurmas
       PreferenciasEntrada RestrProfEntrada Foco UnidadesPorDia TamanhoUnidade}

   in

      % Obtem o numero de aulas
      NumAulas = {Record.width Aulas}
      % Obtem o numero de preferencias
      NumPreferencias = {List.length Preferencias}

      %% PROCEDIMENTO DEVOLVIDO PELA FUNCAO ScriptHorarios %%
      proc {$ Solucao}
         % Implementa o seguinte procedimento auxiliar:
         % Distribuidor.

         % Horarios de inicio das aulas (na ordem dada na tupla Aulas),
         % em unidades, em relacao ao inicio do dia 1.
         Inicios = {FD.tuple inicios NumAulas 0#(UnidadesPorDia*NumDias)}
         % Dias em que ocorrem as aulas (na ordem dada na tupla Aulas).
         Dias = {FD.tuple dias NumAulas 1#NumDias}
         % Duracoes das aulas (na ordem dada na tupla Aulas), em unidades
         Duracoes = {Record.map Aulas fun {$ Aula} Aula.dur end}
         % O conjunto das aulas de um professor e um grupo de exclusao mutua.
         % Assim, para criar os propagadores, podemos juntar as listas de
         % aulas por professor com os grupos de exclusao mutua dados
         % pela entrada.
         GruposExclMut = {List.append Grupos {Dictionary.items Professores}}
         % Lista em que cada valor corresponde a uma preferencia, sendo 0
         % (se a preferencia foi satisfeita) ou o peso da preferencia (caso
         % contrario).
         NaoSatisfeitas

         % DADOS DA SOLUCAO %
         % Dicionario de turmas da solucao
         TurmasSolucao = {Dictionary.new}
         % Lista de preferencias violadas pela solucao
         Violadas
         % Soma das falhas da solucao
         SomaFalhas = {FD.int 0#(NumPreferencias*PesoMaximo)}

         % Procedimento que realiza a distribuicao nos horarios de inicio
         % das aulas contidas no registro HorariosDeInicio.
         % - Para escolher a variavel a ser valorada, utiliza-se uma estrategia
         % "first-fail" um pouco mais "forte", uma vez que os horarios
         % de inicio aparecem aqui na mesma ordem da tupla Aulas, ou seja,
         % a variavel escolhida sera, dentre as que tem dominio de tamanho
         % minimo, a variavel mais "critica", no que diz respeito
         % a interferencia com outras aulas.
         % - Para escolher o valor a ser dado a uma variavel V, utiliza-se uma
         % ideia mais "fraca" de "valor menos restritivo". Seria muito custoso
         % determinar o efeito exato que a escolha de valor para V provoca
         % nos dominios das outras variaveis. Portanto, determina-se apenas
         % o efeito "direto", ou seja, observa-se como ficariam os dominios
         % das variaveis que tem exclusao mutua com V, se escolhessemos um
         % valor para V e aplicassemos as restricoes de exclusao mutua.
         proc {Distribuidor HorariosDeInicio}
            % Recebe:
            % - O indice de uma aula (I1).
            % - V1: um valor possivel para a aula de indice I1.
            % - Uma lista (Indices) com os indices das aulas que tem exclusao
            % mutua com I1.
            % Determina quais seriam os tamanhos dos dominios das aulas que
            % tem exclusao com a aula I1, se esta assumisse o valor V1. Para
            % isso, o unico efeito considerado e a exclusao mutua em si.
            % Devolve um dicionario (representando os efeitos de V1) em que
            % cada chave e um tamanho de dominio e cada elemento e a
            % quantidade de aulas que ficariam com dominio de tal tamanho.
            fun {DicionarioDeRestr I1 V1 Indices}
               % O dicionario que sera devolvido pela funcao
               DicRestr = {Dictionary.new}
               % Devolve o tamanho do dominio de I2 apos a aplicacao da
               % exclusao mutua com I1, caso esta tenha valor V1.
               fun {TamAposRestr I2}
                  FimI1 = V1 + Duracoes.I1
                  DurI2 = Duracoes.I2
               in
                  % Percorre a lista de horarios possiveis de I2, contando
                  % os nao afetados pela exclusao mutua com I1.
                  {List.foldL {FD.reflect.domList Inicios.I2}
                   fun {$ Cont ValorI2}
                      if FimI1 =< ValorI2 orelse ValorI2 + DurI2 =< V1 then
                         Cont+1
                      else Cont end
                   end
                   0
                  }
               end
               % FIM DO PROCEDIMENTO TamAposRestr %
            in
               % Inclui todas as entradas no dicionario...
               {List.forAll Indices
                proc {$ Ind}
                   TamAposRestrIndice = {TamAposRestr Ind}
                   QuantAtual =
                   {Dictionary.condGet DicRestr TamAposRestrIndice 0}
                in
                   {Dictionary.put DicRestr TamAposRestrIndice QuantAtual+1}
                end
               }
               DicRestr
            end
            % FIM DO PROCEDIMENTO DicionarioDeRestr %

            % Recebe dois dicionarios de restricoes, no formato devolvido pela
            % funcao DicionarioDeRestr.
            % Verifica qual e o dicionario que representa um valor menos
            % restritivo, ou seja, que tem maior chave (= tamanho de dominio)
            % minima. Caso ambos tenham a mesma chave minima, verifica-se qual
            % tem menor valor para o elemento correspondente (numero de
            % variaveis com tal tamanho de dominio). Caso continue o empate,
            % as outras chaves sao verificadas, em ordem crescente.
            % Devolve true se RestrV1 representa um valor menos restritivo.
            % Caso contrario, devolve false.
            fun {ComparaRestr RestrV1 RestrV2}
               fun {ComparaTam T1#_ T2#_}
                  T1 < T2
               end
               % Cria listas ordenadas com as entradas dos dicionarios
               ListaOrdV1 = {List.sort {Dictionary.entries RestrV1}
                             ComparaTam}
               ListaOrdV2 = {List.sort {Dictionary.entries RestrV2}
                             ComparaTam}
               % Recebe duas listas, representando as entradas (em ordem
               % crescente de chave) de dicionarios de restricoes.
               % Devolve true se a primeira lista representa um valor menos
               % restritivo que a segunda. Caso contrario, devolve false.
               fun {Melhor L1 L2}
                  case L1 of nil then true
                  [] H1|T1 then
                     case L2 of nil then false
                     [] H2|T2 then
                        if H1.1 == H2.1 then
                           if H1.2 == H2.2 then
                              {Melhor T1 T2}
                           else H1.2 < H2.2 end
                        else H1.1 > H2.1 end
                     end
                  end
               end
            in
               {Melhor ListaOrdV1 ListaOrdV2}
            end
            % FIM DO PROCEDIMENTO ComparaRestr %

            {Space.waitStable}
            % Registro contendo as variaveis ainda nao determinadas.
            Indet = {Record.filter HorariosDeInicio
                     fun {$ I} {FD.reflect.size I} > 1 end
                    }

         in
            case {Record.arity Indet}
            of nil then skip
            [] H|_ then
               % Determina a variavel mais restrita, assim como seu indice.
               IndMaisRestr#MaisRestrita#_ =
               {Record.foldLInd Indet
                fun {$ IndAtual I#Var#TamMin X}
                   if {FD.reflect.size X} < TamMin then
                      IndAtual#X#{FD.reflect.size X}
                   else
                      I#Var#TamMin
                   end
                end
                H#Indet.H#{FD.reflect.size Indet.H}
               }
               % Lista dos valores possiveis para a variavel mais restrita.
               ValoresPossiveis = {FD.reflect.domList MaisRestrita}
               % Linha, na matriz de exclusao mutua, correspondente a variavel
               % mais restrita.
               LinhaMaisRestr = {Array.get MatrizExclMut IndMaisRestr}
               % Variaveis indeterminadas que, pelas restricoes de exclusao
               % mutua, sao afetadas pela variavel mais restrita.
               IndetExclMut =
               {Record.filterInd Indet
                fun {$ IndAtual Var}
                   {Array.get LinhaMaisRestr IndAtual}
                end
               }
               % Determina o valor menos restritivo para a variavel mais
               % restrita
               MenosRestritivo#_ =
               {List.foldL ValoresPossiveis.2
                fun {$ MenosRestrCorrente#DicCorrente Valor}
                   Dicionario =
                   {DicionarioDeRestr IndMaisRestr Valor
                    {Record.arity IndetExclMut}}
                in
                   if {ComparaRestr Dicionario DicCorrente} then
                      Valor#Dicionario
                   else MenosRestrCorrente#DicCorrente end
                end
                ValoresPossiveis.1#({DicionarioDeRestr IndMaisRestr
                                     ValoresPossiveis.1
                                     {Record.arity IndetExclMut}})
               }
            in
               % Finalmente, realiza a distribuicao.
               choice
                  MaisRestrita =: MenosRestritivo
                  {Distribuidor Indet}
               [] MaisRestrita \=: MenosRestritivo
                  {Distribuidor Indet}
               end
            end

         end
         % FIM DO PROCEDIMENTO Distribuidor %

      in
         % Define o formato da solucao
         Solucao = solucao(turmas:TurmasSolucao violadas:Violadas
                           falhas:SomaFalhas)

         %% PROPAGACAO %%

         % Constroi a lista NaoSatisfeitas
         NaoSatisfeitas = {Falhas Inicios Duracoes}
         % Cria um propagador para associar a variavel SomaFalhas ao custo da
         % solucao
         {FD.sum NaoSatisfeitas '=:' SomaFalhas}

         % Relaciona os dias das aulas com os horarios de inicio
         {Record.forAllInd Inicios
          proc {$ Numero Inicio}
             Dias.Numero - 1 =: {FD.divD Inicio UnidadesPorDia}
          end
         }

         % RESTRICAO: cada aula possui uma lista de intervalos em que ela
         % pode comecar.
         {For 1 NumAulas 1
          proc {$ NumAula}
             Inicios.NumAula :: Aulas.NumAula.hor
          end
         }

         % RESTRICAO: alguns professores tem restricoes de horarios.
         {List.forAll RestrProfessores
          proc {$ Restricao}
             HorariosImpossiveis = Restricao.impossiveis
          in
             {List.forAll {Dictionary.get Professores Restricao.prof}
              proc {$ Aula}
                 {List.forAll HorariosImpossiveis
                  proc {$ HI#HF}
                     {FD.disj Inicios.Aula =<: HI - Duracoes.Aula
                      Inicios.Aula >=: HF 1}
                  end
                 }
              end
             }
          end
         }

         % RESTRICAO: distancias (minima e maxima), em dias, entre as aulas
         % de uma mesma turma.
         % OUTRA RESTRICAO: eliminacao de simetrias (vide comentarios no
         % codigo abaixo).
         {List.forAll {Dictionary.items Turmas}
          proc {$ Turma}
             {List.forAllTail Turma.aulas
              proc {$ Tail}
                 case Tail
                 of Prim|Resto then
                    {List.forAll Resto
                     proc {$ Outra}
                        Distancia in
                        % Eliminacao de simetrias: se duas aulas de uma mesma
                        % turma sao identicas (tem a mesma duracao e a mesma
                        % lista de horarios possiveis), pode-se criar a
                        % restricao de que uma delas deve ocorrer antes.
                        if (Aulas.Prim.dur == Aulas.Outra.dur) andthen
                           (Aulas.Prim.hor == Aulas.Outra.hor) then
                           Inicios.Prim + Duracoes.Prim =<: Inicios.Outra
                        end
                        Distancia::(Turma.dist.min)#(Turma.dist.max)
                        {FD.distance Dias.Prim Dias.Outra '=:' Distancia}
                     end
                    }
                 else skip end
              end
             }
          end
         }

         % RESTRICAO: grupos de exclusao mutua entre aulas.
         {Schedule.taskIntervals GruposExclMut Inicios Duracoes}

         % RESTRICAO: cada aula possui uma lista de aulas que nao podem
         % coincidir com ela.
         {For 1 NumAulas 1
          proc {$ Numero}
             ExclMut = Aulas.Numero.exclMut
          in
             {List.forAll ExclMut
              proc {$ NumOutra}
                 {FD.disjoint Inicios.Numero Duracoes.Numero
                  Inicios.NumOutra Duracoes.NumOutra}
              end
             }
          end
         }

         % RESTRICAO: existem turmas cujas aulas nao podem comecar no mesmo
         % horario do dia.
         {List.forAll {Dictionary.items Turmas}
          proc {$ Turma}
             if Turma.horVaria then
                IniciosNoDia =
                {List.map Turma.aulas
                 fun {$ NumAula} {FD.modI Inicios.NumAula UnidadesPorDia} end
                }
             in
                {FD.distinct IniciosNoDia}
             end
          end
         }

         % DISTRIBUICAO %

         % FORMA DE DISTRIBUICAO: tentar primeiro as solucoes que satisfazem
         % as preferencias (se desejado pelo usuario)
         if Foco == otima then
            {List.forAll NaoSatisfeitas
             proc {$ NaoSatisf}
                {Space.waitStable}
                choice NaoSatisf =: 0 [] NaoSatisf >: 0 end
             end
            }
         end

         % FORMA DE DISTRIBUICAO: nos horarios de inicio, com estrategia
         % "variavel mais restrita" e "valor menos restritivo".
         {Distribuidor Inicios}

         % PREPARA UMA ESTRUTURA MAIS CLARA PARA A SOLUCAO

         % Dicionario de turmas
         {List.forAll {Dictionary.entries Turmas}
          proc {$ IdTurma#Turma}
             % Lista de aulas desta turma
             AulasSolucao =
             {List.map Turma.aulas
              fun {$ NumAula}
                 HorMinutos =
                 (Inicios.NumAula mod UnidadesPorDia) * TamanhoUnidade
              in
                 % Formato de uma aula, na solucao
                 aula(dia:Dias.NumAula
                      hor:(HorMinutos div 60)#(HorMinutos mod 60)
                      dur:(Duracoes.NumAula*TamanhoUnidade))
              end
             }
          in
             {Dictionary.put TurmasSolucao IdTurma AulasSolucao}
          end
         }

         % Lista de preferencias violadas
         if Preferencias \= nil then
            {List.foldL Preferencias
             fun {$ ListaAtual Pref}
                Indesejados = Pref.indesejados
                Professor = Pref.prof
                % Lista de aulas que violam esta preferencia
                AulasQueViolam =
                {List.foldL {Dictionary.get Professores Professor}
                 fun {$ Atual Aula}
                    % true, se esta aula viola a preferencia.
                    % Caso contrario, false.
                    Viola =
                    {List.foldL Indesejados
                     fun {$ ResultAtual HI#HF}
                        ResultAtual orelse
                        (Inicios.Aula < HF andthen
                         Inicios.Aula + Duracoes.Aula > HI)
                     end
                     false $
                    }
                 in
                    % Se a aula viola a preferencia, ela e posta na lista
                    if Viola then
                       TotalMinutos =
                       (Inicios.Aula mod UnidadesPorDia)*TamanhoUnidade
                       Acrescentar =
                       aula(turma:Aulas.Aula.turma dia:Dias.Aula
                            hor:((TotalMinutos div 60)#(TotalMinutos mod 60))
                            dur:(Aulas.Aula.dur*TamanhoUnidade))
                    in
                       Acrescentar|Atual
                    else
                       Atual
                    end
                 end
                 nil $
                }
             in
                % Se nenhuma aula viola, esta preferencia nao e posta na lista
                % de preferencias violadas
                if AulasQueViolam == nil then
                   ListaAtual
                else
                   {List.append ListaAtual
                    [violada(prof:Professor aulas:AulasQueViolam
                             peso:Pref.peso)]}
                end
             end
             nil Violadas
            }
         else Violadas = nil end

      end

   end
   % FIM DA FUNCAO ScriptHorarios %

   % Cria propagadores para a restricao de que SolucaoNova deve ser melhor
   % que SolucaoVelha
   proc {Ordem SolucaoVelha SolucaoNova}
      % Observe que uma condicao necessaria para a solucao nova ser melhor
      % e que a solucao antiga nao seja "perfeita".
      SolucaoVelha.falhas >: 0
      SolucaoVelha.falhas >: SolucaoNova.falhas
   end

end