% INTERFACE ENTRE O NUCLEO DO PROGRAMA E O MODULO Horarios %

functor

import
   Search
   Nucleo

export
   CriaResolvedor

define
   % Sao implementados os seguintes procedimentos:
   % CriaObjetosResolvedores, CriaResolvedor.


   % Recebe:
   % - Os dados de entrada de um problema.
   % - Um dicionario (TurmasPorProf), indexado por professores, contendo a
   % lista de turmas (apenas identificadores) de cada professor.
   % Verifica se existem professores cujas aulas nao interferem com as aulas
   % de outros professores, ou seja, professores cujas turmas formam
   % problemas independentes.
   % Devolve uma lista de N objetos Search.object, sendo que (N-1) objetos
   % representam problemas independentes com um professor cada e o objeto
   % restante representa o problema principal sem os professores independentes.
   fun {CriaObjetosResolvedores NumDias Turmas GruposExclMut Preferencias
        RestrProf Foco TurmasPorProf}
      % Implementa o seguinte procedimento auxiliar:
      % ProfIndependente.

      % Devolve true se nao existir exclusao mutua entre alguma turma do
      % professor Prof e uma turma de outro professor (levando em conta tanto
      % as lista de exclusao mutua de cada disciplina quanto os grupos de
      % exclusao mutua). Caso contrario, devolve false.
      fun {ProfIndependente Prof}
         % Verifica as listas de exclusao mutua de cada turma e devolve true
         % se e somente se nao existir exclusao mutua entre alguma turma do
         % professor Prof e uma turma de outro professor.
         {List.foldL {Dictionary.items Turmas}
          fun {$ ResultAtual Turma}
             if Turma.exclMut \= nil then
                ResultAtual andthen
                if Turma.prof == Prof then
                   % Verifica se todas as turmas sao do professor Prof
                   {List.foldL Turma.exclMut
                    fun {$ Atual IdOutra}
                       Atual andthen
                       {Dictionary.get Turmas IdOutra}.prof == Prof
                    end
                    true $
                   }
                else
                   % Verifica se todas as turmas sao de outros professores
                   {List.foldL Turma.exclMut
                    fun {$ Atual IdOutra}
                       Atual andthen
                       {Dictionary.get Turmas IdOutra}.prof \= Prof
                    end
                    true $
                   }
                end
             else ResultAtual end
          end
          true $
         }

         andthen

         % Verifica cada grupo de exclusao mutua e devolve true se, para todo
         % grupo, ou nenhuma das turmas e do professor Prof ou todas elas sao.
         {List.foldL GruposExclMut
          fun {$ ResultAtual Grupo}
             % O professor da ultima turma deste grupo (a escolha da ultima
             % e arbitraria).
             ProfUltima = {Dictionary.get Turmas {List.last Grupo}}.prof
          in
             ResultAtual andthen
             % Compara os professores de todas as turmas com Prof.
             if ProfUltima == Prof then
                {List.foldL Grupo
                 fun {$ Atual IdTurma}
                    Atual andthen
                    {Dictionary.get Turmas IdTurma}.prof == Prof
                 end
                 true $
                }
             % Verifica se todos os professores sao diferentes de Prof.
             else
                {List.foldL Grupo
                 fun {$ Atual IdTurma}
                    Atual andthen
                    {Dictionary.get Turmas IdTurma}.prof \= Prof
                 end
                 true $
                }
             end
          end
          true $
         }
      end
      %% FIM DO PROCEDIMENTO ProfIndependente %%

      % Lista de professores tal que, para cada um, temos:
      % nenhuma das turmas do professor tem exclusao mutua com alguma
      % turma de outro professor.
      ProfessoresIndependentes
      NumProfsIndep
      % Lista de objetos Search.object para os problemas independentes que
      % contem apenas um professor cada.
      ObjsProfsIndep
      % Lista de objetos para todos os problemas independentes.
      ObjetosResolvedores
   in
      % Obtem a lista de professores independentes.
      ProfessoresIndependentes =
      {List.filter {Dictionary.keys TurmasPorProf} ProfIndependente}
      % Numero de professores independentes
      NumProfsIndep = {List.length ProfessoresIndependentes}

      ObjsProfsIndep = {List.make NumProfsIndep}

      % Cria um objeto resolvedor para cada problema independente relacionado
      % a um professor.
      {List.forAllInd ProfessoresIndependentes
       proc {$ Ind Prof}
          % Cria o dicionario de turmas do subproblema.
          TurmasDeste = {Dictionary.new}
          % A lista de preferencias do subproblema.
          PrefsDeste =
          {List.filter Preferencias
           fun {$ Pref}
              Pref.prof == Prof
           end
          }
          % A lista de restricoes do subproblema.
          RestrDeste = {List.filter RestrProf
                        fun {$ Restr} Restr.prof == Prof end
                       }
       in
          % Preenche o dicionario de turmas do subproblema.
          {List.forAll {Dictionary.get TurmasPorProf Prof}
           proc {$ IdTurma}
              {Dictionary.put TurmasDeste IdTurma
               {Dictionary.get Turmas IdTurma}}
           end
          }
          % Cria um objeto resolvedor para o subproblema.
          {List.nth ObjsProfsIndep Ind} =
          {New Search.object
           script({Nucleo.scriptHorarios NumDias TurmasDeste nil PrefsDeste
                   RestrDeste Foco} Nucleo.ordem rcd:1)}
       end
      }

      % Cria o ultimo objeto resolvedor (para o problema restante, excluidos
      % os professores que representam problemas independentes), se for
      % necessario.

      % Caso nao-trivial: existem professores independentes.
      if ProfessoresIndependentes \= nil then
         % Dicionario de turmas do ultimo problema. Inicialmente, contem todas
         % as turmas do problema original. Porem, serao excluidas as turmas
         % dos professores independentes.
         TurmasUltimo = {Dictionary.clone Turmas}
         % Grupos de exclusao mutua do ultimo problema.
         GruposUltimo =
         {List.filter GruposExclMut
          fun {$ Primeira|_}
             ProfPrimeira = {Dictionary.get Turmas Primeira}.prof
          in
             {Bool.'not' {List.member ProfPrimeira ProfessoresIndependentes}}
          end
         }
         % Preferencias do ultimo problema.
         PreferenciasUltimo =
         {List.filter Preferencias
          fun {$ Pref}
             {Bool.'not' {List.member Pref.prof ProfessoresIndependentes}}
          end
         }
         % Restricoes de horarios dos professores do ultimo problema.
         RestrUltimo =
         {List.filter RestrProf
          fun {$ Restr}
             {Bool.'not' {List.member Restr.prof ProfessoresIndependentes}}
          end
         }
      in
         % Subtrai, do dicionario de turmas do ultimo problema, as turmas
         % dos professores independentes.
         {List.forAll ProfessoresIndependentes
          proc {$ Prof}
             {List.forAll {Dictionary.get TurmasPorProf Prof}
              proc {$ IdTurma}
                 {Dictionary.remove TurmasUltimo IdTurma}
              end
             }
          end
         }
         % Se o ultimo problema nao contem turmas, ele, obviamente, nao precisa
         % ser resolvido.
         if {Dictionary.isEmpty TurmasUltimo} then
            ObjetosResolvedores = ObjsProfsIndep
         else
            % Cria o objeto Search.object para o ultimo problema.
            UltimoObj =
            {New Search.object
             script({Nucleo.scriptHorarios NumDias TurmasUltimo GruposUltimo
                     PreferenciasUltimo RestrUltimo Foco} Nucleo.ordem rcd:1)}
         in
            % Define a lista com os objetos relativos a todos os problemas.
            ObjetosResolvedores = {List.append ObjsProfsIndep [UltimoObj]}
         end
      else  % Caso trivial: nao existem professores independentes.
         ObjetosResolvedores =
         [ {New Search.object
            script({Nucleo.scriptHorarios NumDias Turmas GruposExclMut
                    Preferencias RestrProf Foco} Nucleo.ordem rcd:1)}
         ]
      end

      ObjetosResolvedores
   end
   %% FIM DO PROCEDIMENTO CriaObjetosResolvedores. %%

   % Recebe os dados de entrada do problema.
   % Devolve um registro na forma:
   % resolvedor(proximaSolucao:ProximaSolucao pausa:Pausa
   % solucaoOtima:SolucaoOtima)
   % Onde:
   % - ProximaSolucao e uma funcao que busca uma solucao melhor que a ultima
   % encontrada e devolve a nova solucao numa lista de um elemento (ou nil,
   % caso nao exista solucao melhor que a ultima encontrada).
   % - Pausa e uma funcao que interrompe a busca atual (caso haja uma busca em
   % execucao).
   % - SolucaoOtima e uma funcao que busca a melhor solucao possivel para o
   % problema e devolve esta solucao numa lista de um elemento (ou nil, caso
   % nao exista solucao).
   fun {CriaResolvedor NumDias Turmas GruposExclMut Preferencias RestrProf
        Foco}
      % Implementa os seguintes procedimentos auxiliares:
      % UneSolucoes, AtualizaSolucao, ProximaSolucao, Pausa, SolucaoOtima.

      % Lista com os objetos resolvedores que serao criados para os problemas
      % independentes.
      ObjetosResolvedores
      % Numero de objetos resolvedores criados.
      NumObjResolvedores

      % Quando a funcao ProximaSolucao for chamada pela primeira vez, esta
      % variavel sera vinculada a unit.
      ProximaSolucaoJaChamada

      % Numero do problema que deve ser resolvido na proxima chamada a
      % ProximaSolucao. Os problemas sao resolvidos na ordem dada pela lista
      % ObjetosResolvedores.
      ProblemaAtual = {NewCell 1}
      % A solucao atual do problema (ultima solucao devolvida pela funcao
      % ProximaSolucao).
      SolucaoAtual = {NewCell nil}
      % Dicionario, indexado por professores, contendo a lista das turmas
      % (apenas identificadores) de cada professor.
      TurmasPorProf = {Dictionary.new}

      % Recebe: uma lista contendo uma solucao para cada subproblema
      % independente.
      % Devolve: a solucao correspondente para o problema original.
      fun {UneSolucoes ListaDeSolucoes}
         % A solucao correspondente (em uma lista de um elemento)
         % para o problema principal (nil, se nao existir solucao).
         SolEmLista =
         {List.foldL ListaDeSolucoes
          fun {$ SolAtual Sol}
             % Solucoes nil sao totalmente ignoradas.
             if Sol == nil then SolAtual
             else
                % Se esta e a primeira solucao diferente de nil, entao a
                % solucao para o problema original, por enquanto, e esta.
                if SolAtual == nil then Sol
                else  % As solucoes de fato (a partir das listas de 1 elemento)
                   SolDeFato = Sol.1
                   SolAtualDeFato = SolAtual.1
                in
                   % Acrescenta as turmas desta solucao aa solucao atual do
                   % problema original.
                   {List.forAll {Dictionary.entries SolAtualDeFato.turmas}
                    proc {$ IdTurma#Turma}
                       {Dictionary.put SolDeFato.turmas IdTurma Turma}
                    end
                   }
                   [solucao(turmas:SolDeFato.turmas
                            violadas:{List.append SolDeFato.violadas
                                      SolAtualDeFato.violadas}
                            falhas:SolDeFato.falhas+SolAtualDeFato.falhas)]
                end
             end
          end
          nil $
         }
      in
         if SolEmLista \= nil then
            SolEmLista.1
         else nil end
      end
      %% FIM DO PROCEDIMENTO UneSolucoes. %%

      % Recebe:
      % - Uma solucao para o problema original (SolCorrente).
      % - Uma nova solucao para um dos subproblemas (NovaSolucao).
      % Devolve uma solucao atualizada para o problema original, incorporando
      % a nova solucao do subproblema.
      fun {AtualizaSolucao SolCorrente NovaSolucao}
         solucao(turmas:T violadas:V falhas:_) = SolCorrente
         solucao(turmas:TNS violadas:VNS falhas:FNS) = NovaSolucao
         % Professores do subproblema
         ProfsNovaSolucao =
         {List.foldL {Dictionary.entries TurmasPorProf}
          fun {$ ListaAtual Prof#(Turma1|_)}
             if
                % Verifica se este professor pertence ao subproblema
                {List.foldL {Dictionary.keys TNS}
                 fun {$ ResultAtual IdTurma}
                    ResultAtual orelse IdTurma == Turma1
                 end
                 false $
                }
             then
                {List.append ListaAtual [Prof] $}
             else ListaAtual end
          end
          nil $
         }
         % Lista de preferencias violadas, na solucao atualizada.
         Violadas
         % Custo da solucao corrente, sem contar as preferencias dos
         % professores do subproblema.
         Custo
      in
         % Sobrescreve as turmas da solucao corrente que foram
         % recalculadas na nova solucao do subproblema.
         {List.forAll {Dictionary.entries TNS}
          proc {$ IdTurma#Turma}
             {Dictionary.put T IdTurma Turma}
          end
         }
         % Recria a lista de preferencias violadas, substituindo
         % as preferencias do professor da solucao atual.
         Violadas =
         {List.append VNS
          if V == nil then
             Custo = 0
             nil
          else
             % Cria uma lista com as preferencias violadas (na solucao
             % corrente) dos professores que nao pertencem ao subproblema.
             % O valor de Custo tambem e calculado.
             {List.foldL V
              fun {$ CustoAtual#ListaAtual Violada}
                 if {List.member Violada.prof ProfsNovaSolucao} then
                    CustoAtual#ListaAtual
                 else
                    (CustoAtual+Violada.peso)#{List.append
                                               ListaAtual
                                               [Violada]}
                 end
              end
              0#nil Custo#$
             }
          end
          $
         }
         solucao(turmas:T violadas:Violadas falhas:(Custo+FNS))
      end
      %% FIM DO PROCEDIMENTO AtualizaSolucao. %%

      % Devolve a proxima solucao do problema, numa lista de um elemento (nil,
      % caso nao exista proxima solucao).
      fun {ProximaSolucao}
         % Implementa a seguinte funcao auxiliar:
         % ProxSol.

         % Devolve a proxima solucao do problema atual. Se nao existir, passa
         % para o proximo problema (e assim por diante).
         fun {ProxSol NumAtual}
            % Verifica se todos os subproblemas ja foram resolvidos.
            if NumAtual > NumObjResolvedores then
               % Atualiza o numero do problema atual (neste caso, nao sera
               % resolvido mais nenhum problema).
               {Cell.assign ProblemaAtual NumAtual}
               nil
            else  % Busca a proxima solucao do proximo problema.
               Sol = {{List.nth ObjetosResolvedores NumAtual} next($)}
            in
               % Se nao ha solucao para o problema atual, passa-se para
               % o proximo problema.
               if Sol == nil then {ProxSol NumAtual+1}
               else
                  % Atualiza o numero do problema atual.
                  {Cell.assign ProblemaAtual NumAtual}
                  Sol.1
               end
            end
         end
         %% FIM DO PROCEDIMENTO ProxSol. %%

      in
         % Verifica se todos os subproblemas ja foram resolvidos.
         if {Cell.access ProblemaAtual} > NumObjResolvedores then
            nil
         else
            % Verifica se esta e a primeira chamada a funcao ProximaSolucao.
            if {Value.isDet ProximaSolucaoJaChamada} then
               % Se a funcao ProximaSolucao ja foi chamada e a solucao atual
               % e nil, entao o problema nao tem solucao.
               if {Cell.access SolucaoAtual} == nil then
                  nil
               else
                  % Encontra a proxima solucao.
                  Solucao = {ProxSol {Cell.access ProblemaAtual}}
               in
                  if Solucao == nil then
                     {Cell.assign SolucaoAtual nil}
                     nil
                  else
                     % Atualiza a solucao atual do problema original, levando
                     % em conta a nova solucao do subproblema atual.
                     {Cell.assign SolucaoAtual
                      {AtualizaSolucao {Cell.access SolucaoAtual} Solucao}
                     }
                     [{Cell.access SolucaoAtual}]
                  end
               end
            else % Esta e a primeira chamada a funcao ProximaSolucao
               % Encontra a proxima solucao de cada subproblema.
               ProximasSolucoes = {List.map ObjetosResolvedores
                                   fun {$ Objeto} {Objeto next($)} end
                                  }
               % true, se a primeira solucao de algum subproblema e nil.
               % Caso contrario, false.
               AlgumaNil =
               {List.foldL ProximasSolucoes
                fun {$ ResultAtual Sol}
                   ResultAtual orelse (Sol == nil)
                end
                false $
               }
            in
               % Vincula a variavel que indica que a funcao ProximaSolucao ja
               % foi chamada.
               ProximaSolucaoJaChamada = unit
               % Se a primeira solucao de algum subproblema e nil, entao o
               % problema original nao tem solucao.
               if AlgumaNil then nil
               else
                  % Obtem a primeira solucao do problema original, a partir
                  % da primeira solucao de cada subproblema.
                  {Cell.assign SolucaoAtual {UneSolucoes ProximasSolucoes}}
                  [{Cell.access SolucaoAtual}]
               end
            end
         end
      end
      %% FIM DO PROCEDIMENTO ProximaSolucao. %%

      % Interrompe a busca atual (se alguma busca estiver sendo executada).
      proc {Pausa}
         {List.forAll ObjetosResolvedores
          proc {$ ObjResolv}
             {ObjResolv stop}
          end
         }
      end
      %% FIM DO PROCEDIMENTO Pausa. %%

      % Devolve uma solucao otima para o problema (numa lista de um elemento),
      % partindo da ultima solucao encontrada. Devolve nil, se nao existir
      % solucao melhor que a solucao corrente.
      fun {SolucaoOtima}
         % Obtem a solucao otima de cada subproblema.
         SolucoesOtimas = {List.map ObjetosResolvedores
                           fun {$ Objeto} {Objeto last($)} end
                          }
         % TodasNil e true se e somente se nenhum dos subproblemas tem solucao.
         % AlgumaNil e true se e somente se pelo menos um dos subproblemas nao
         % tem solucao.
         TodasNil#AlgumaNil =
         {List.foldL SolucoesOtimas
          fun {$ RR#RA Sol}
             EstaENil = (Sol == nil)
          in
             (RR andthen EstaENil)#(RA orelse EstaENil)
          end
          true#false $
         }
      in
         % Observe que talvez a funcao ProximaSolucao ja tenha sido chamada e,
         % portanto, e necessario levar em conta a solucao corrente.
         if (AlgumaNil andthen {Cell.access SolucaoAtual} == nil) orelse
            TodasNil then
            {Cell.assign SolucaoAtual nil}
            nil
         else
            if AlgumaNil then  % A solucao corrente nao e nil.
               % Atualiza a solucao corrente com as solucoes otimas encontradas
               % para os subproblemas.
               {List.forAll SolucoesOtimas
                proc {$ Sol}
                   if Sol \= nil then
                      { Cell.assign SolucaoAtual
                        {AtualizaSolucao {Cell.access SolucaoAtual} Sol.1} }
                   end
                end
               }
            else
               % Foram encontradas solucoes otimas para todos os subproblemas.
               % Portanto, a funcao UneSolucoes e capaz de construir uma
               % solucao otima para o problema original.
               {Cell.assign SolucaoAtual {UneSolucoes SolucoesOtimas}}
            end
            [{Cell.access SolucaoAtual}]
         end
      end
      %% FIM DO PROCEDIMENTO SolucaoOtima. %%

   in
      % Obtem as turmas de cada professor
      {ForAll {Dictionary.entries Turmas}
       proc {$ IdTurmaEntrada#TurmaEntrada}
          % Professor dessa turma.
          Prof = TurmaEntrada.prof
          % Lista atual de turmas desse professor (pode ser nil).
          TurmasDoProf = {Dictionary.condGet TurmasPorProf Prof nil}
       in
          {Dictionary.put TurmasPorProf Prof {List.append TurmasDoProf
                                              [IdTurmaEntrada]}}
       end
      }

      % Cria um objeto Search.object para cada subproblema independente
      % encontrado.
      ObjetosResolvedores = {CriaObjetosResolvedores NumDias Turmas
                             GruposExclMut Preferencias RestrProf Foco
                             TurmasPorProf}
      NumObjResolvedores = {List.length ObjetosResolvedores}
      resolvedor(proximaSolucao:ProximaSolucao pausa:Pausa
                 solucaoOtima:SolucaoOtima)
   end
   %% FIM DO PROCEDIMENTO CriaResolvedor. %%

end